from typing import Tuple, List, Union

import torch
from torchvision.models.detection import fasterrcnn_resnet50_fpn


class FasterRCNNModule(torch.nn.Module):
    """Module for Faster R-CNN Model with ResNet-50 backbone pre-trained on Coco dataset from PyTorch

    Usage:
    ```python
    >>> model = FasterRCNNModule()
    >>> output = model()
    ```

    **Parameters:**

    """

    def __init__(self):
        super(FasterRCNNModule, self).__init__()
        self.model = fasterrcnn_resnet50_fpn(pretrained=True)

    def forward(
        self, images: Union[torch.Tensor, List[torch.Tensor]]
    ) -> Union[Tuple[torch.Tensor, torch.Tensor, torch.Tensor], List[torch.Tensor]]:
        """Forward pass on this Module only requires a list of dynamic images of shape [C, H, W].
        To work with a Triton Server which does not encode Python list objects, we enable a
        `torch.Tensor` input that we can trace a model compatible with Triton inference server.

        Both inputs are traceable with this model.

        If input is list of `torch.Tensor` then the output will be a List of torch Tensors moddable by
        3 where every 3 entries corresponds to one image.


        * ** images ** - List of tensor images of shape [C, H, W]
        """
        if isinstance(images, torch.Tensor):
            predictions = self.model([images])
            return (
                predictions[0]["boxes"],
                predictions[0]["labels"],
                predictions[0]["scores"],
            )
        else:
            predictions = self.model(images)
            return self._dict2tensor(predictions)

    def _dict2tensor(self, predictions):
        prediction_outputs = []
        for p in predictions:
            if "masks" in p.keys():
                prediction_outputs += [p["boxes"], p["labels"], p["scores"], p["masks"]]
            prediction_outputs += [p["boxes"], p["labels"], p["scores"]]
        return prediction_outputs



from pathlib import Path
from typing import Optional, Callable

import torch
from torchvision.datasets.vision import VisionDataset
from torchvision.transforms import ConvertImageDtype
from torchvision.io import read_image


class ImageDataset(VisionDataset):
    def __init__(
        self, root: str, transforms: Optional[Callable] = ConvertImageDtype(torch.float)
    ):
        super(ImageDataset, self).__init__(root, transforms=transforms)
        self.dir_path = Path(root)
        self.image_paths = [
            p for p in self.dir_path.iterdir() if not p.name.startswith(".")
        ]

    def __len__(self):
        return len(self.image_paths)

    def __getitem__(self, idx):
        img = read_image(self.image_paths[idx].absolute().as_posix())
        tensor_image = self.transforms(img)
        return tensor_image



from typing import Tuple
import logging

import torch

from tritonclient.http import (
    InferenceServerClient,
    InferInput,
    InferRequestedOutput,
    InferResult,
)

logger = logging.getLogger(__name__)


class FastNNClient:
    """FastNN client class for pushing requests to the Triton Inference Server


    Usage:
    ```python
    >>> client = FastNNClient(url="127.0.0.1:8000", model_name="distilbert-squad", model_version="1")
    >>> client.request(batch=batch)
    ```

    **Parameters:**

    * **url** - String url of Triton Inference Server. Defaults to 127.0.0.1:8000
    * **model_name** - String name of model in `model_repository` directory
    * **model_version** - String model version name
    """

    def __init__(
        self,
        url: str = "127.0.0.1:8000",
        model_name: str = "distilbert-squad",
        model_version: str = "1",
        verbose: bool = False,
    ):
        self.url = url
        self.model_name = model_name
        self.model_version = model_version

        self.triton_client = InferenceServerClient(url=url, verbose=verbose)
        self.model_metadata = self.triton_client.get_model_metadata(
            model_name=model_name, model_version=model_version
        )
        self.model_config = self.triton_client.get_model_config(
            model_name=model_name, model_version=model_version
        )

    def request(
        self, batch: Tuple[torch.Tensor], binary_data: bool = False
    ) -> InferResult:
        """Runs a request with the `batch` input that can be generated from a FastNN `Processor`

        * **batch** - Tuple of torch tensors, typically batch inputs from a dataloader
        """
        inputs_metadata = self.model_metadata["inputs"]
        outputs_metadata = self.model_metadata["outputs"]

        # Assert batch input matches triton model metadata
        assert len(batch) == len(inputs_metadata)

        inputs = []
        for i, metadata in enumerate(inputs_metadata):
            inp = InferInput(
                metadata["name"], tuple(batch[i].shape), metadata["datatype"]
            )
            inp.set_data_from_numpy(batch[i].cpu().numpy(), binary_data=binary_data)
            inputs.append(inp)

        outputs = []
        for i, metadata in enumerate(outputs_metadata):
            out = InferRequestedOutput(metadata["name"], binary_data=binary_data)
            outputs.append(out)

        response = self.triton_client.infer(
            model_name=self.model_name,
            model_version=self.model_version,
            inputs=inputs,
            outputs=outputs,
        )

        return response



from abc import ABC, abstractmethod


class Processor(ABC):
    """Abstract class for `Processor` objects in fastnn for the purpose of processing tensor objects to and from human interfacing data."""

    @abstractmethod
    def process(self):
        """ Returns a dataset at the minimum """
        raise NotImplementedError("Please Implement this method")

    @abstractmethod
    def process_batch(self):
        """ """
        raise NotImplementedError("Please Implement this method")

    @abstractmethod
    def process_output(self):
        """ """
        raise NotImplementedError("Please Implement this method")

    @abstractmethod
    def process_output_batch(self):
        """ """
        raise NotImplementedError("Please Implement this method")



import logging
from typing import Optional, List, Union, Tuple, Callable

import torch
from torch.utils.data import DataLoader, Dataset

from torchvision.transforms import ConvertImageDtype

from PIL import Image, ImageDraw
from PIL import ImageFont
import numpy as np


from fastnn.utils.cv import ImageDataset


from fastnn.processors.base_processing import Processor


logger = logging.getLogger(__name__)


class ObjectDetectionProcessor(Processor):
    """Object Detection processor dealing with image files or 3xHxW formatted images and boxes, scores, labels out processing.
    Since most resizing and padding transforms are done by the object detection models in PyTorch, datasets and dataloaders willl
    generate batches of images as lists.

    Usage:
    ```python
    >>> processor = ObjectDetectionProcessor()
    >>> processor.process(file_paths=["file_path.png"])

    **Parameters:**
    * **label_strings** - List of strings that specify label strings with index as key for this specific processor

    ```
    """

    def __init__(self, label_strings: List[str]):
        self.label_strings = label_strings

    def process(
        self,
        dir_path: str,
        transforms: Optional[Callable] = ConvertImageDtype(torch.float),
    ) -> Dataset:
        """Generate torch `Dataset` object from list of file paths or image Tensors.
        This provides clear tensor input representations for compatible models.

        Returns a Dataset

        * **dir_path** - String path to directory of images you'd like to process
        """

        dataset = ImageDataset(root=dir_path, transforms=transforms)

        return dataset

    def process_batch(
        self,
        dir_path: str,
        transforms: Optional[Callable] = ConvertImageDtype(torch.float),
        mini_batch_size: int = 8,
        use_gpu: bool = False,
    ) -> DataLoader:
        """Generate torch `Dataloader` object from data directory path.
        This provides clear tensor input representations for compatible models.

        Returns a `Dataloader`

        * **dir_path** - String path to directory of images you'd like to process
        * **mini_batch_size** - Batch size for inference
        * **use_gpu** - Bool for using gpu or cpu. If set True but no gpu devices available, model will default to using cpu
        """

        if use_gpu:
            if torch.cuda.is_available():
                device = torch.device("cuda")
            else:
                logger.info("GPU not available")
                device = torch.device("cpu")
        else:
            device = torch.device("cpu")

        dataset = self.process(dir_path=dir_path, transforms=transforms)

        # Instead of a tensor batch, the lambda collate_fn will provide a list batch
        dataloader = DataLoader(
            dataset,
            batch_size=mini_batch_size,
            collate_fn=lambda x: [[t.to(device) for t in self._od_collate_fn(x)]],
        )

        return dataloader

    def process_output(
        self,
    ):
        pass

    def process_output_batch(
        self, outputs: List[List[torch.Tensor]], dataset: Dataset
    ) -> List[List[Tuple[torch.Tensor, np.array]]]:
        """Process output of object detection model into human legible results.
        Outputs from `FasterRCNNModule`


        Returns batched results of list of list of tuples containing boxed images in tensor and numpy format

        * **outputs** - List of batch output tensors from a model's forward pass
        * **dataset** - Corresponding dataset with originial images matched with model outputs

        """
        # Labeled Images
        results = []

        for idx, out in enumerate(outputs):
            labeled_images = []
            for label_idx in range(1, len(out), 3):
                labels = [self.label_strings[o] for o in out[label_idx]]

                unique_labels = set(labels)
                label_colors_map = {}
                for label in unique_labels:
                    label_colors_map[label] = tuple(
                        np.random.choice(range(256), size=3)
                    )

                label_colors = [label_colors_map[label] for label in labels]

                output_tensor, output_numpy = self.draw_bounding_boxes(
                    ConvertImageDtype(torch.uint8)(
                        dataset[idx * (len(out) // 3) + label_idx // 3]
                    ),
                    out[label_idx - 1],
                    labels=labels,
                    colors=label_colors,
                )
                labeled_images.append((output_tensor, output_numpy))
            results.append(labeled_images)

        return results

    def _od_collate_fn(self, data):
        """Custom collate fn to output dynamic image batches without same-dim requirements via. `stack`.
        This is not technically a "correct" collate_fn for most of torch's vision models. Should be wrapped as a list
        in the lambda collate fn.
        """
        data = [img for img in data]
        return data

    @torch.no_grad()
    def draw_bounding_boxes(
        self,
        image: torch.Tensor,
        boxes: torch.Tensor,
        labels: Optional[List[str]] = None,
        colors: Optional[List[Union[str, Tuple[int, int, int]]]] = None,
        width: int = 1,
        font: Optional[str] = "arial.ttf",
        font_size: int = 10,
    ) -> Tuple[torch.Tensor, np.array]:

        """
        Added and modified from TorchVision utils.
        Draws bounding boxes on given image.
        The values of the input image should be uint8 between 0 and 255.
        Args:
            image (Tensor): Tensor of shape (C x H x W)
            bboxes (Tensor): Tensor of size (N, 4) containing bounding boxes in (xmin, ymin, xmax, ymax) format. Note that
                the boxes are absolute coordinates with respect to the image. In other words: `0 <= xmin < xmax < W` and
                `0 <= ymin < ymax < H`.
            labels (List[str]): List containing the labels of bounding boxes.
            colors (List[Union[str, Tuple[int, int, int]]]): List containing the colors of bounding boxes. The colors can
                be represented as `str` or `Tuple[int, int, int]`.
            width (int): Width of bounding box.
            font (str): A filename containing a TrueType font. If the file is not found in this filename, the loader may
                also search in other directories, such as the `fonts/` directory on Windows or `/Library/Fonts/`,
                `/System/Library/Fonts/` and `~/Library/Fonts/` on macOS.
            font_size (int): The requested font size in points.
        """

        if not isinstance(image, torch.Tensor):
            raise TypeError(f"Tensor expected, got {type(image)}")
        elif image.dtype != torch.uint8:
            raise ValueError(f"Tensor uint8 expected, got {image.dtype}")
        elif image.dim() != 3:
            raise ValueError("Pass individual images, not batches")

        ndarr = image.permute(1, 2, 0).numpy()
        img_to_draw = Image.fromarray(ndarr)

        img_boxes = boxes.to(torch.int64).tolist()

        draw = ImageDraw.Draw(img_to_draw)

        pixel_ratio = max(1, (max(ndarr.shape[0], ndarr.shape[1]) // 1000))

        for i, bbox in enumerate(img_boxes):
            color = None if colors is None else colors[i]
            draw.rectangle(bbox, width=width * pixel_ratio, outline=color)

            if labels is not None:
                txt_font = (
                    ImageFont.load_default()
                    if font is None
                    else ImageFont.truetype(font=font, size=font_size * pixel_ratio)
                )
                draw.text((bbox[0], bbox[1]), labels[i], fill=color, font=txt_font)

        return torch.from_numpy(np.array(img_to_draw)).permute(2, 0, 1), np.array(
            img_to_draw
        )



import logging
from typing import List, Tuple
from collections import OrderedDict

import torch
from torch.utils.data import DataLoader, Dataset
from torch.utils.data.dataloader import default_collate

from transformers import (
    SquadExample,
    squad_convert_examples_to_features,
    AutoTokenizer,
)

from transformers.data.processors.squad import SquadFeatures, SquadResult

from fastnn.utils.qa.squad_metrics import (
    compute_predictions_logits,
)

from fastnn.processors.base_processing import Processor


logger = logging.getLogger(__name__)


class TransformersQAProcessor(Processor):
    """Question Answering Data Processor. Use this class to generate tensor inputs from human legible text/string data.
    This class can be used with a majority of the Bert architecture transformer models with the span-based extractive,
    Question Answering predictive head from Hugging Face.

    Usage:
    ```python
    >>> processor = TRansformersQAProcessor(model_name_or_path="distilbert-base-cased-distilled-squad")
    >>> processor.process(query=["string"], context["string"])

    **Parameters:**

    * **model_name_or_path** - String defining HF question answering model/tokenizer's name
    ```
    """

    def __init__(
        self, model_name_or_path: str = "distilbert-base-cased-distilled-squad"
    ):
        self.tokenizer = AutoTokenizer.from_pretrained(
            model_name_or_path
        )  # Can't use fast tokenizer yet for QA `use_fast=True`

    def process(
        self,
        query: List[str],
        context: List[str],
        max_seq_length: int = 512,
        doc_stride: int = 128,
        max_query_length: int = 64,
    ) -> Tuple[List[SquadExample], List[SquadFeatures], Dataset]:
        """Generate torch `Dataset` object from query/context string pairs using specified tokenizer from HF.
        This provides clear tensor input representations for compatible models.

        Returns a tuple `Dataset` and matching `SquadFeatures`

        * **query** - List of query strings, must be same length as `context`
        * **context** - List of context strings, must be same length as `query`
        * **max_seq_length** - Maximum context token length. Check model configs to see max sequence length the model was trained with
        * **doc_stride** - Number of token strides to take when splitting up context into chunks of size `max_seq_length`
        * **max_query_length** - Maximum token length for queries
        """
        examples = self._generate_squad_examples(query=query, context=context)
        features, dataset = squad_convert_examples_to_features(
            examples,
            self.tokenizer,
            max_seq_length=max_seq_length,
            doc_stride=doc_stride,
            max_query_length=max_query_length,
            is_training=False,
            return_dataset="pt",
            threads=1,
        )

        return examples, features, dataset

    def process_batch(
        self,
        query: List,
        context: List,
        mini_batch_size: int = 8,
        max_seq_length: int = 512,
        doc_stride: int = 128,
        max_query_length: int = 64,
        use_gpu: bool = False,
    ) -> Tuple[List[SquadExample], List[SquadFeatures], DataLoader]:
        """Generate torch `DataLoader` object from query/context string pairs using specified tokenizer from HF.
        This provides clear tensor input representations for compatible models in an easy to use batch

        Returns a tuple of (List[`SquadExample`], List[`SquadFeatures`], `DataLoader`)

        * **query** - List of query strings, must be same length as `context`
        * **context** - List of context strings, must be same length as `query`
        * **mini_batch_size** - Batch size for inference
        * **max_seq_length** - Maximum context token length. Check model configs to see max sequence length the model was trained with
        * **doc_stride** - Number of token strides to take when splitting up context into chunks of size `max_seq_length`
        * **max_query_length** - Maximum token length for queries
        * **use_gpu** - Bool for using gpu or cpu. If set True but no gpu devices available, model will default to using cpu
        """

        if use_gpu:
            if torch.cuda.is_available():
                device = torch.device("cuda")
            else:
                logger.info("GPU not available")
                device = torch.device("cpu")
        else:
            device = torch.device("cpu")

        examples, features, dataset = self.process(
            query=query,
            context=context,
            max_seq_length=max_seq_length,
            doc_stride=doc_stride,
            max_query_length=max_query_length,
        )

        dataloader = DataLoader(
            dataset,
            batch_size=mini_batch_size,
            collate_fn=lambda x: [t.to(device) for t in self._qa_collate_fn(x)],
        )

        return examples, features, dataloader

    def process_output(
        self,
        outputs: List[Tuple[torch.Tensor, torch.Tensor, torch.Tensor]],
        examples: List[SquadExample],
        features: List[SquadFeatures],
        n_best_size: int = 5,
        max_answer_length: int = 10,
        do_lower_case: bool = False,
        verbose_logging: bool = False,
        version_2_with_negative: bool = False,
        null_score_diff_threshold: float = 0.0,
    ):
        pass

    def process_output_batch(
        self,
        outputs: List[Tuple[torch.Tensor, torch.Tensor, torch.Tensor]],
        examples: List[SquadExample],
        features: List[SquadFeatures],
        n_best_size: int = 5,
        max_answer_length: int = 64,
        do_lower_case: bool = False,
        verbose_logging: bool = False,
        version_2_with_negative: bool = False,
        null_score_diff_threshold: float = 0.0,
    ) -> Tuple[OrderedDict, OrderedDict]:
        """Process output of Transformers QA model into human legible results.

        * **outputs** - List of batch output tensors from a model's forward pass
        * **examples** - List of `SquadExample` objects for each original context/query pair used as input. This is returned from the built-in `process()` or `process_batch()` methods
        * **features** - List of `SquadFeature` objects for each context/query pair over the original doc_stride lengths. This is also returned from the built-in `process()` or `process_batch()` methods
        * **n_best_size** - Number of top n results you want
        * **max_answer_length** - Maximum token length for answers that are returned
        * **do_lower_case** - Set as `True` if using uncased QA models
        * **verbose_logging** - Set True if you want prediction verbose loggings
        * **version_2_with_negative** - Set as True if using QA model with SQUAD2.0
        * **null_score_diff_threshold** - Threshold for predicting null(no answer) in Squad 2.0 Model.  Default is 0.0.  Raise this if you want fewer null answers
        """

        # Generate results per example query
        all_results: List[SquadResult] = []
        for output in outputs:
            example_indices = output[2]
            for i, example_index in enumerate(example_indices):
                start_logits = self._to_list(output[0][i])
                end_logits = self._to_list(output[1][i])
                eval_feature = features[example_index[0].item()]
                unique_id = int(eval_feature.unique_id)
                result = SquadResult(unique_id, start_logits, end_logits)
                all_results.append(result)

        # Compute predictions based off logits on a per example basis
        answers, n_best = compute_predictions_logits(
            all_examples=examples,
            all_features=features,
            all_results=all_results,
            n_best_size=n_best_size,
            max_answer_length=max_answer_length,
            do_lower_case=do_lower_case,
            verbose_logging=verbose_logging,
            version_2_with_negative=version_2_with_negative,
            null_score_diff_threshold=null_score_diff_threshold,
            tokenizer=self.tokenizer,
        )

        return answers, n_best

    def _qa_collate_fn(self, data):
        batch = default_collate(data)
        # Generate same batch dims for scalars to address future batch inferencing
        batch[3].unsqueeze_(1)
        batch[4].unsqueeze_(1)
        return batch

    def _to_list(self, tensor: torch.Tensor):
        return tensor.detach().cpu().tolist()

    def _generate_squad_examples(
        self, query: List[str], context: List[str]
    ) -> List[SquadExample]:
        """Generate HF Squad Example objects with query/context pairs"""
        assert len(query) == len(context)
        examples = []
        title = "qa"
        is_impossible = False
        answer_text = None
        start_position_character = None
        answers = ["answer"]
        for idx, (q, c) in enumerate(zip(query, context)):
            example = SquadExample(
                qas_id=str(idx),
                question_text=q,
                context_text=c,
                answer_text=answer_text,
                start_position_character=start_position_character,
                title=title,
                is_impossible=is_impossible,
                answers=answers,
            )
            examples.append(example)
        return examples



import logging
from typing import Union
from pathlib import Path

import torch
from torch.utils.data import DataLoader

logger = logging.getLogger(__name__)


class TorchScriptExporter:
    """Exporter Class for turning torch models into torch script models. Requires a torch Module and corresponding `DataLoader` populated with complete data samples.
    This means the `DataLoader` has atleast one full batch sample that can be used as input to the torch Module in order to run `torch.jit.trace()`

    Usage:
    ```python
    >>> exporter = TorchScriptExporter(model=BertQAModule(...), dataloader=dataloader, use_gpu=True)
    >>> exporter.export()
    ```

    **Parameters:**

    * **model** - Torch model/module, preferably from `fastnn.nn`
    * **dataloader** - Torch `DataLoader` that corresponds to the model
    * **use_gpu** - Bool for using gpu or cpu. If set True but no gpu devices available, model will default to using cpu

    """

    def __init__(
        self, model: torch.nn.Module, dataloader: DataLoader, use_gpu: bool = False
    ):
        if use_gpu:
            if torch.cuda.is_available():
                device = torch.device("cuda")
                logger.info(f"Torch model set to device {device}")
            else:
                device = torch.device("cpu")
                logger.info(f"GPU not available...device set to {device}")
        else:
            device = torch.device("cpu")
            logger.info(f"Torch model set to device {device}")

        self.model = model
        self.model.eval()
        self.model.to(device)
        self.dataloader = dataloader
        self.torchscript_model = None

    def export(self) -> torch.jit.ScriptModule:
        """Traces pytorch model and returns `ScriptModule` model"""
        batch_input = next(iter(self.dataloader))
        self.torchscript_model = torch.jit.trace(self.model, tuple(batch_input))
        return self.torchscript_model

    def serialize(self, file_path: Union[Path, str]):
        """Serialize and save model

        * **file_path** - String file path to save serialized torchscript model
        """
        if isinstance(file_path, str):
            file_path = Path(file_path)

        if self.torchscript_model:
            file_path.parent.mkdir(parents=True, exist_ok=True)
            self.torchscript_model.save(str(file_path))



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['fastnn',
 'fastnn.nn',
 'fastnn.processors',
 'fastnn.processors.cv',
 'fastnn.processors.nlp',
 'fastnn.utils',
 'fastnn.utils.qa']

package_data = \
{'': ['*']}

install_requires = \
['coremltools>=4.0,<5.0',
 'jupyter>=1.0.0,<2.0.0',
 'jupyterlab>=2.2.9,<3.0.0',
 'matplotlib>=3.3.3,<4.0.0',
 'transformers>=4.0.0,<5.0.0',
 'wget>=3.2,<4.0']

extras_require = \
{'docs': ['mkdocs>=1.1.2,<2.0.0',
          'mkdocs-material>=6.1.5,<7.0.0',
          'mkautodoc>=0.1.0,<0.2.0'],
 'torch': ['torch>=1.0.0,<2.0.0', 'torchvision<1.0.0']}

setup_kwargs = {
    'name': 'fastnn',
    'version': '0.0.4',
    'description': 'A python library and framework for fast neural network computations.',
    'long_description': None,
    'author': 'Andrew Chang',
    'author_email': 'aychang995@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import Tuple

import torch
from transformers import AutoModelForQuestionAnswering


class BertQAModule(torch.nn.Module):
    """Module for Bert-based Question Answering models from Transformers

    Usage:
    ```python
    >>> model = BertQAModule(model_name_or_path="distilbert-base-cased-distilled-squad")
    >>> output = model(*batch)
    ```

    **Parameters:**

    * **model_name_or_path** - String key of Transformer's pre-trained model hosted in Hugging Face's Model Repository
    """

    def __init__(
        self, model_name_or_path: str = "distilbert-base-cased-distilled-squad"
    ):
        super(BertQAModule, self).__init__()
        self.model = AutoModelForQuestionAnswering.from_pretrained(
            model_name_or_path, torchscript=True
        )

    def forward(
        self, input_ids, attention_mask, token_type_ids, example_indices, *args
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Forward pass on this QA module requires input_ids, attention_mask, and example_indices.
        The `example_indices` parameter is required specifically for span-based extractive question anwering model outputs
        in order to retain the structure of inputs within the final start and end logit calculations.

        * **input_ids** - Tensor generated from FastNN `Processor` class
        * **attention_mask** - Tensor generated from FastNN `Processor` class
        * **token_type_ids** - Tensor generated from FastNN `Processor` class
        * **example_indices** - Tensor generated from FastNN `Processor` class
        * ** &ast;args ** - args available to make abstraction easier for variant `Processor` classes
        """
        start_logits, end_logits = self.model(
            input_ids=input_ids, attention_mask=attention_mask
        )
        return start_logits, end_logits, example_indices
